
// Generic fd part of comgt

// © 2020 Jonas Bechtel
// Licensed under the Apache License, Version 2.0
//     see file <prj>/meta/NOTICE


#ifndef GENERIC_FD
#define GENERIC_FD

//#include <stdint.h>

#ifndef NOCODE
//#define _POSIX_C_SOURCE 202001L
#define _XOPEN_SOURCE 700
//#define _DEFAULT_SOURCE

#include <unistd.h> // write
#include <string.h> // strlen
#include <stdio.h>
#include <stdint.h>
#include <poll.h>
#include <errno.h>
#include <stdlib.h> // posix_openpt etc.
#include <termio.h> // struct termio
#include <sys/types.h>
#include <sys/socket.h>
#include "fcntl-compat.h" // fcntl, O_RDWR
#define NOCODE
#include "comgt.c"
#undef NOCODE

#define xstr(s) str(s)
#define str(s) #s

#endif // #ifndef NOCODE


struct generic_fd {
	int  fd;
	char type; // 0, 'c'om, 'p'ty, 'f'ile, 'r'ecycled fd, 's'ocketpair  ['c' does not work yet]
	uint8_t  copied:1; // \x00 if normal, \x01 if copied  // little bool
			// 0xfe is reserved for future use
	uint8_t  pair:1; // 1 for type 's', 0 otherwise
	uint16_t name; // "pointer" into the gfdnames array.
	//char name[3];
};

#define MAXFD 10

#ifdef NOCODE
extern
#endif // #ifdef NOCODE
struct generic_fd gfds[MAXFD];


#ifndef NOCODE
// Convention: zero-terminated strings. No zero-length string allowed -> this marks end of array.
#define MAXGFDNAMETOTALSIZE 2048
char gfdnames[MAXGFDNAMETOTALSIZE];
#endif // #ifndef NOCODE

void init_gfd(void)
#ifndef NOCODE
{
	int i;
	for (i = 0; i < MAXFD; i++) {
		gfds[i].type = 0;
		gfds[i].name = 0xffff;
	}
	gfdnames[0] = 0;
}
#endif // #ifndef NOCODE
;

void destroy_gfd(void)
#ifndef NOCODE
{
	// TODO: destroy all gfds.
	
	// unlinking the pty links:
	
	int i;
	for (i = 0; i < MAXFD; i++) {
		if (gfds[i].copied & 1) continue;
		if (gfds[i].type == 0)  continue;
		if (gfds[i].fd == 0)    continue; // This is be done somewhere else in a libc function
		close (gfds[i].fd);
		
		if (gfds[i].type != 'p') continue;
		if (gfds[i].name == 0xffff) continue;
		
		unlink(&(gfdnames[gfds[i].name]));
	}
}
#endif // #ifndef NOCODE
;


int getfd(char* script, off_t* pc)
#ifndef NOCODE
{
    skipspaces_noglobal(script, pc);
	if(script[*pc] != '!')
		return(-1);
	(*pc)++;
	unsigned int idx = getvalue();
	if (idx >= MAXFD)
		serror("Maximum number of gfds is " str(MAXFD), 7);
	//printf("getfd(): gfd=%d, fd=%d\n", idx, gfds[idx].fd);
	return idx;
}
#endif // #ifndef NOCODE
;

// Returns the number of not yet written characters
//		(0 = all fine)
// or -1 to -9 in case of scripting errors
//		(-1 = Gfd not open)
//		(-2 = Other side of gfd not open)
// or (0-(10+errno)) in case of system errors
int writefd(int gfd, char* string)
#ifndef NOCODE
{
	if (gfds[gfd].type == 0) return -1;
	if (gfds[gfd].type == 'p') {
		// Check whether we've been connected.
		struct pollfd pret = {.fd = gfds[gfd].fd, .events = 0};
		int ret = poll(&pret, 1, 0);
		if (ret > 0 && pret.revents & POLLHUP)
			return -2;
	}
	
	size_t len = strlen(string);
	ssize_t written = write(gfds[gfd].fd, string, len);
	if (written < 0) return 0 - 10 - errno;
	return len - written;
}
#endif // #ifndef NOCODE
;




#ifndef NOCODE
static uint16_t gfd_findend(uint16_t starthint) {
	bool zero = true;
	while (starthint < MAXGFDNAMETOTALSIZE) {
		// Declare it char-by-char - hopefully get's well optimized
		char c = gfdnames[starthint++];
		if (zero && c == 0) break;
		zero = (c == 0);
	}
	return starthint - 1;
}
#endif // #ifndef NOCODE


// Returns
//		(0 = all fine)
// or -1 to -9 in case of scripting errors
//		(-1 = gfd does not exist)
//		(-2 = gfd has no name)
//		(-3 = name of active pty must not be cleared (locked for unlink))
// returns new write position in newmem (if set).
int gfd_clrname(int gfd, uint16_t* newmem)
#ifndef NOCODE
{
	if (gfds[gfd].type == 0)      return -1;
	if (gfds[gfd].name == 0xffff) return -2;
	if (gfds[gfd].type == 'p')    return -3;
	
	if (gfds[gfd].copied & 1 == 1) {
		// Check whether this string is unique
		int i;
		for (i = 0; i < MAXFD; i++) {
			if (i == gfd) continue;
			if (gfds[i].type != gfds[gfd].type) continue; // catches type == 0
			if (gfds[i].name != gfds[gfd].name) continue;
			
			// Another gfd used this name exist -> no memory operation needed. But we need to find the end of memory, as we promised.
			if (newmem)
				*newmem = gfd_findend(gfds[gfd].name);
			gfds[gfd].name = 0xffff;
			return 0;
		}
	}
	
	uint16_t wrpos = gfds[gfd].name;
	gfds[gfd].name = 0xffff; // The actual clear operation
	
	// Do memmove and search(\0\0) in one loop
	uint16_t delta = strlen(&(gfdnames[wrpos]));
	if (delta == 0) return -2; // This must not happen
	uint16_t rdpos = wrpos + delta;
	bool zero = false;
	while (rdpos < MAXGFDNAMETOTALSIZE) {
		// Declare it char-by-char - hopefully get's well optimized
		char c = gfdnames[rdpos++];
		gfdnames[wrpos++] = c;
		if (zero && c == 0) break;
		if (c == 0) zero = true;
	}
	if (newmem) *newmem = wrpos - 1;
	return 0;
}
#endif
;

// name is not needed after finish -> can be freed.
// Returns
//		(0 = all fine)
// or -1 to -9 in case of scripting errors
//		(-1 = not enough space)
//		(-3 = name of active pty must not be altered (locked for unlink))
//				But you can copy the gfd and alter the copy's name.
int gfd_setname(int gfd, const char* name)
#ifndef NOCODE
{
	if (gfds[gfd].type == 'p') return -3;
	uint16_t newmem = 0xffff;
	int res = gfd_clrname(gfd, &newmem);
	if (res != 0)
		newmem = gfd_findend(0);
	
	uint16_t space = MAXGFDNAMETOTALSIZE - newmem;
	uint16_t len = strlen(name);
	if (space < len)
		return -1;
	strcpy(&(gfdnames[newmem]), name);
	gfds[gfd].name = newmem;
	return 0;
}
#endif // #ifndef NOCODE
;


const char* gfd_getname(int gfd)
#ifndef NOCODE
{
	if (gfds[gfd].name == 0xffff)
		return ""; // = empty zero-terminated string = pointer to a 0
	return &(gfdnames[gfds[gfd].name]);
}
#endif // #ifndef NOCODE
;


// Close gfd
//		 0 = all fine
//		-1 = gfd does not exist
// or (0-(10+errno)) in case of system errors
int gfd_close(int gfd)
#ifndef NOCODE
{
	if(gfds[gfd].type==0) return -1;
	
	bool symlink_ok = true;
	if(     gfds[gfd].type=='p'
		&& (gfds[gfd].copied & 1 == 0)
		&&  gfds[gfd].name != 0xffff) {
		
		int res = unlink(&(gfdnames[gfds[gfd].name]));
		symlink_ok = (res == 0);
	}
	
	if (gfds[gfd].copied & 1 == 0) {
		// Check whether there are copies around
		int i;
		for (i = 0; i < MAXFD; i++) {
			if (i == gfd) continue;
			if (gfds[i].type != gfds[gfd].type) continue;
			if (gfds[i].fd   != gfds[gfd].fd  ) continue;
			
			// Found one copy. Transfer "original" state to it.
			gfds[gfd].copied |=  1; // set   copied-flag
			gfds[i  ].copied &= ~1; // clear copied-flag
			break;
		}
	}
	
	gfd_clrname(gfd, NULL);
	
    int res = close(gfds[gfd].fd);
    if(res < 0) return 0 - (10 + errno);
    gfds[gfd].type = 0; // Mark as free
    return 0;
}
#endif // #ifndef NOCODE
;



// Open fd rfd as !gfd
//		 0 = all fine
//		-1 = gfd1 does not exist
//		-2 = gfd2 already in use
// or (0-(10+errno)) in case of system errors
int gfd_copy(int gfd1, int gfd2)
#ifndef NOCODE
{
	if (gfds[gfd1].type == 0) return -1;
	if (gfds[gfd2].type != 0) return -2;
	
	gfds[gfd2].type   = gfds[gfd1].type;
	gfds[gfd2].fd     = gfds[gfd1].fd;
	gfds[gfd2].copied = 1;
	gfds[gfd2].pair   = gfds[gfd1].pair;
	gfds[gfd2].name   = gfds[gfd1].name;
	return 0;
}
#endif // #ifndef NOCODE
;


// Open fd rfd as !gfd
//		 0 = all fine
//		-1 = gfd already in use
// or (0-(10+errno)) in case of system errors
int recyclefd(int gfd, int rfd)
#ifndef NOCODE
{
	if (gfds[gfd].type != 0) return -1;
	int res = fcntl(rfd, F_GETFD);
	if (res < 0) return 0 - 10 - errno;
	
	gfds[gfd].type   = 'r';
	gfds[gfd].fd     = rfd;
	gfds[gfd].copied = 0;
	gfds[gfd].pair   = 0;
	return 0;
}
#endif // #ifndef NOCODE
;


// Returns 0
//		0 = all fine
// or -1 to -9 in case of scripting errors
//		(-1 = Gfd already open)
// or (0-(10+errno)) in case of system errors
int openfile(int gfd, const char* filename)
#ifndef NOCODE
{
	if (gfds[gfd].type != 0)
		return -1;
	int fd = open(filename, O_RDWR);
	if (fd < 0)
		return 0 - (10 + errno);
	gfds[gfd].fd     = fd;
	gfds[gfd].type   = 'f';
	gfds[gfd].copied = 0;
	gfds[gfd].pair   = 0;
	return 0;
}
#endif // #ifndef NOCODE
;



// Returns 0
//		0 = all fine
// or -1 to -9 in case of scripting errors
//		(-1 = gfd1 already open)
//		(-2 = gfd2 already open or equal to gfd1)
// or (0-(10+errno)) in case of system errors
int gfd_opensocketpair(int gfd1, int gfd2)
#ifndef NOCODE
{
	if (gfds[gfd1].type != 0)
		return -1;
	if (gfds[gfd2].type != 0)
		return -2;
	if (gfd1 == gfd2)
		return -2;
	int fds[2];
	int res = socketpair(AF_UNIX, SOCK_STREAM, 0, fds);
	if (res < 0)
		return 0 - (10 + errno);
	gfds[gfd1].fd     = fds[0];
	gfds[gfd1].type   = 's';
	gfds[gfd1].copied = 0;
	gfds[gfd1].pair   = 1;
	gfds[gfd2].fd     = fds[1];
	gfds[gfd2].type   = 's';
	gfds[gfd2].copied = 0;
	gfds[gfd2].pair   = 1;
	return 0;
}
#endif // #ifndef NOCODE
;





// Returns 0
//		0 = all fine
// or -1 to -9 in case of scripting errors
//		(-1 = Gfd already open)
// or (0-(10+errno)) in case of system errors (of openpt/grantpt/unlockpt)
int openpty(int gfd)
#ifndef NOCODE
{
	if (gfds[gfd].type != 0) return -1;
	int fd = posix_openpt(O_RDWR | O_NOCTTY);
	if (fd < 0) return 0-(10+errno);
	int res = grantpt(fd);
	if (fd < 0) {close(fd); return 0-(10+errno);}
	
	struct termio tty;
	ioctl(fd, TCGETA, &tty);
	//printf("Terminal setting flags: %x\n", tty.c_lflag);
	tty.c_lflag    &= ~(ICANON|ECHO|ECHOE|ECHOK|ECHONL);
	ioctl(fd, TCSETA, &tty);
	
	
	res = unlockpt(fd);
	if (fd < 0) {close(fd); return 0-(10+errno);}
	gfds[gfd].fd = fd;
	
	
		/*struct pollfd pret = {.fd = fd, .events = 0};
		int ret = poll(&pret, 1, 0);
		if (ret > 0)
			printf("Polled flags = %d (POLLHUP=%d)\n", pret.revents,POLLHUP);
		fflush (stdout);*/
		
	// Trick: open and close pty -> POLLHUP then has a meaning
    // Man kann rausfinden, ob ein Client verbunden ist, indem man sich kurz selbst als Client verbindet und wieder trennt und dann auf HUP-Signale hört. https://stackoverflow.com/questions/3486491/how-can-i-detect-when-someone-opens-the-slave-side-of-a-pty-pseudo-terminal-in
    // https://stackoverflow.com/a/3490197
	res = open(ptsname(fd), O_RDWR);
	if (res >= 0)
		close(res);
	
		/*ret = poll(&pret, 1, 0);
		if (ret > 0)
			printf("Polled flags = %d (POLLHUP=%d)\n", pret.revents,POLLHUP);
		fflush (stdout);*/
	
	
	gfds[gfd].type   = 'p';
	gfds[gfd].copied = 0;
	gfds[gfd].pair   = 0;
	if (gfds[gfd].name != 0xffff) {
		res = symlink(ptsname(fd), &(gfdnames[gfds[gfd].name]));
		if (res < 0) {
			int myerrno = errno;
			fprintf(stderr, "Couldn't create symlink @'%s' (%d = %s)!\n", &(gfdnames[gfds[gfd].name]), myerrno, strerror(errno));
			gfds[gfd].name = 0xffff;
			return 0-(10+myerrno);
		}
	}
	
	
	return 0;
	
	// TODO: stty, geht das von hier aus?
	// 4004:4:bf:a00:3:1c:7f:15:4:0:1:0:11:13:1a:0:12:f:17:16:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0
	// bzw.
/*
speed 38400 baud; line = 0;
min = 1; time = 0;
-brkint -icrnl -imaxbel iutf8
-opost
-isig -icanon -iexten -echo -echoe -echok
*/
    // Das ist schon ganz gut, allerdings werden die Zeichen aus irgendeinem Grund gleichzeitig zurückgespiegelt.
	// (Update: der Kopie-Effekt tritt dann auf, wenn schon ein Teil geprintet wurde, während auf der Gegenseite noch gar keine Verbindung existierte.)
    // Noch besser ist übrigens "stty sane"
    // Siehe auch http://www.mit.edu/afs.new/athena/astaff/reference/olc-stock/stock_answers.real/c/c_terminal_echo
    // man 4 ioctl_tty, man 1 stty
    
}
#endif
;


#endif // #ifndef GENERIC_FD


